# SPDX-FileCopyrightText: StorPool <support@storpool.com>
# SPDX-License-Identifier: BSD-2-Clause
"""Routines for checking out the OS-specific packaging files."""

from __future__ import annotations

import pathlib
import re
import subprocess
import sys
import tempfile
import typing

from . import defs
from . import util


if typing.TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Final


RE_DEB_ORIG: Final = re.compile(
    r""" ^
    (?P<name>
        [A-Za-z0-9.-]+
    )
    _
    (?P<version>
        .+
    )
    \.orig
    (?:
        -
        (?P<component>
            [A-Za-z0-9-]+
        )
    )?
    \.tar
    (?P<ext>
        \.
        (?: gz | bz2 | xz )
    )
    $ """,
    re.X,
)


def deb_orig_components(
    cfg: defs.Config, archives: list[pathlib.Path]
) -> tuple[pathlib.Path, dict[str, pathlib.Path]]:
    """Parse the names of the Debian *.orig{-comp}?.tar.* files."""
    cfg.diag("- looking for Debian orig source tarballs")
    components: Final = {}
    for path in archives:
        cfg.diag(f"  - checking {path}")
        data = RE_DEB_ORIG.match(path.name)
        if not data:
            sys.exit(f"Unexpected source tarball {path}")

        comp = data.groupdict().get("component", "")
        if comp is None:
            comp = ""
        cfg.diag(f"    - got component name {comp!r}")
        components[comp] = path

    cfg.diag(f"  - got {sorted(components)}")
    if "" not in components:
        sys.exit(f"Could not recognize a single orig tarball for {cfg.pkgname} among {archives}")

    return (
        components[""],
        {name: path for name, path in components.items() if name},
    )


def extract_and_move_into(
    cfg: defs.Config,
    archive: pathlib.Path,
    destdir: pathlib.Path,
    omit: Iterable[str] | None = None,
) -> None:
    """Extract an archive, move all its files into the specified directory."""
    cfg.diag(f"- extracting {archive} into {destdir}")
    if not destdir.is_dir():
        cfg.diag(f"  - creating {destdir}")
        destdir.mkdir(mode=0o755)

    omit_names: Final = set(omit) if omit is not None else set()

    with tempfile.TemporaryDirectory(dir=destdir.parent, prefix=destdir.name + ".") as tempdname:
        tempd: Final = pathlib.Path(tempdname)
        cfg.diag(f"  - extracting {archive} into {tempd}")
        cfg.check_call(
            ["tar", "-xaf", archive],
            cwd=tempd,
        )

        extracted: Final = list(tempd.iterdir())
        if len(extracted) != 1 or not extracted[0].is_dir():
            sys.exit(
                f"Expected a single directory after extracting "
                f"{archive} into {tempd}, got {sorted(extracted)}"
            )
        temp_up: Final = extracted[0]
        cfg.diag(f"  - extracted into {temp_up}")

        cfg.diag(f"  - moving everything from {temp_up} to {destdir}")
        for src in temp_up.iterdir():
            dst = destdir / src.name
            if src.name in omit_names:
                cfg.diag("    - skipping {src}")
                continue

            try:
                src.rename(dst)
            except OSError as err:
                sys.exit(f"Could not rename {src} to {dst}: {err}")


def link_recursive(
    cfg: defs.Config,
    srcdir: pathlib.Path,
    dstdir: pathlib.Path,
    *,
    exist_ok: bool,
) -> None:
    """Recursively link a directory tree somewhere else."""
    cfg.diag(f"- linking everything under {srcdir} into {dstdir}")
    for src in srcdir.rglob("*"):
        dst = dstdir / src.relative_to(srcdir)
        if src.is_dir():
            dst.mkdir(mode=src.resolve().stat().st_mode, exist_ok=exist_ok)
        else:
            if exist_ok and dst.exists():
                dst.unlink()
            src.resolve().link_to(dst)


def checkout_debian(
    cfg: defs.Config,
    archives: list[pathlib.Path],
    distfiles: list[pathlib.Path],
) -> None:
    """Check out a Debian package in place."""
    assert cfg.pkgname is not None  # noqa: S101  # mypy needs this
    assert cfg.path.work is not None  # noqa: S101  # mypy needs this

    if archives:
        main_tarball, components = deb_orig_components(cfg, archives)

        extract_and_move_into(cfg, main_tarball, cfg.path.work, omit=["debian"])

        for name, path in components.items():
            compdir = cfg.path.work / name
            cfg.check_call(["rm", "-rf", "--", compdir])
            extract_and_move_into(cfg, path, compdir)

        cfg.diag(f"- linking the distfiles into {cfg.path.work.parent}")
        for src in distfiles:
            dst = cfg.path.work.parent / src.name
            if dst.exists():
                dst.unlink()
            cfg.diag(f"  - {src} -> {dst}")
            src.resolve().link_to(dst)

        debsrc: Final = cfg.path.package / "debian"
        debdst: Final = cfg.path.work / "debian"
        debdst.mkdir(mode=debsrc.resolve().stat().st_mode)
        link_recursive(cfg, debsrc, debdst, exist_ok=False)
    else:
        cfg.diag("- a native package, not extracting anything")
        link_recursive(cfg, cfg.path.package, cfg.path.work, exist_ok=True)


def checkout_centos(
    cfg: defs.Config,
    _archives: list[pathlib.Path],
    distfiles: list[pathlib.Path],
) -> None:
    """Check out a package for CentOS building."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    assert cfg.path.work is not None  # noqa: S101  # mypy needs this
    srcdir: Final = cfg.path.work / "src"
    cfg.diag(f"- recreating {srcdir}")
    cfg.check_call(["rm", "-rf", "--", srcdir])
    srcdir.mkdir(mode=0o755)

    cfg.diag(f"- linking the distfiles into {srcdir}")
    for src in distfiles:
        dst = srcdir / src.name
        cfg.diag(f"  - {src} -> {dst}")
        src.resolve().link_to(dst)

    cdir: Final = cfg.path.package / cfg.branch.flavor
    link_recursive(cfg, cdir, srcdir, exist_ok=False)


def find_source_archives(
    cfg: defs.Config, package: defs.Package
) -> tuple[list[pathlib.Path], list[pathlib.Path]]:
    """Check out the Git branch and find the archive paths."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    cfg.diag("- parsing the checksums file")
    cksums_file: Final = cfg.path.project / package.checksums[cfg.branch.flavor]
    distfiles: Final = sorted(cfg.path.dist / fname for fname in util.parse_checksums(cksums_file))
    if not distfiles:
        # No distfiles and parse_checksums() did not raise an error, so
        # this is a native package. Right?
        return [], []

    missing: Final = [path for path in distfiles if not path.is_file()]
    if missing:
        sys.exit(f"Missing source archives: {missing}")

    archives: Final = sorted(
        path
        for path in distfiles
        if any(path.name.endswith(suffix) for suffix in defs.ARCHIVE_SUFFIXES)
    )
    if not archives:
        sys.exit(f"Could not recognize any archives in {cksums_file}")

    return archives, distfiles


def checkout_overlays(cfg: defs.Config, package: defs.Package) -> None:
    """Link any overlay files if present."""
    if not package.overlays:
        return
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    assert cfg.branch.flavor == "debian"  # noqa: S101  # mypy needs this
    assert cfg.path.work is not None  # noqa: S101  # mypy needs this

    for path in package.overlays:
        link_recursive(cfg, path, cfg.path.work, exist_ok=True)


def cmd_checkout(cfg: defs.Config) -> None:
    """Check a single package out for building."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    assert cfg.pkgname is not None  # noqa: S101  # mypy needs this
    assert cfg.path.work is not None  # noqa: S101  # mypy needs this

    cfg.diag(f"Check out {cfg.pkgname} for use in {cfg.path.package}")
    package: Final = util.get_package_for_branch(cfg)
    archives, distfiles = find_source_archives(cfg, package)

    subprocess.check_call(["rm", "-rf", "--", cfg.path.work], env=cfg.subenv)
    cfg.path.work.mkdir(parents=True)

    if cfg.branch.flavor == "debian":
        checkout_debian(cfg, archives, distfiles)
    else:
        checkout_centos(cfg, archives, distfiles)

    checkout_overlays(cfg, package)
