# SPDX-FileCopyrightText: StorPool <support@storpool.com>
# SPDX-License-Identifier: BSD-2-Clause
"""Definitions for the StorPool package helper."""

from __future__ import annotations

import dataclasses
import os
import subprocess
from typing import TYPE_CHECKING, List, Sequence, Union

import cfg_diag
from sp_variant import variant


if TYPE_CHECKING:
    import pathlib
    from typing import IO, Final

    Path = os.PathLike[str]
    Popen = subprocess.Popen[str]
else:
    Path = os.PathLike
    Popen = subprocess.Popen

PathSeq = Sequence[Union[Path, str]]
PathList = List[Union[Path, str]]


VERSION: Final = "1.1.0"

DISTROS: Final = ("centos", "debian")

ARCHIVE_SUFFIXES: Final = (".tar.gz", ".tar.bz2", ".tar.xz", ".tgz")


@dataclasses.dataclass(frozen=True)
class Branch:
    """The definition of a build branch."""

    name: str
    flavor: str
    repo_id: str

    def chroot_name(self) -> str:
        """Determine the name of the build chroot environment."""
        last: Final = self.name.split("/", maxsplit=1)[-1]
        if self.flavor == "centos":
            return f"storpool-c{last}-x86_64"
        return f"sp-{last}-amd64"


@dataclasses.dataclass(frozen=True)
class Paths:
    """Paths used during the building of the packages."""

    project: pathlib.Path
    package: pathlib.Path
    dist: pathlib.Path
    work: pathlib.Path | None
    repo: pathlib.Path | None
    sprepo: pathlib.Path | None
    remote_sppkg: pathlib.Path | None


@dataclasses.dataclass(frozen=True)
class Config(cfg_diag.ConfigDiag):
    """Runtime configuration for the StorPool package build tool."""

    branch: Branch | None
    branch_only: bool
    mirror: str | None
    path: Paths
    pkgname: str | None
    subenv: dict[str, str]

    def check_call(
        self,
        cmd: PathSeq,
        *,
        shell: bool = False,
        env: dict[str, str] | None = None,
        cwd: pathlib.Path | None = None,
    ) -> None:
        """Run a child process in a UTF-8 locale."""
        subprocess.check_call(
            cmd,
            shell=shell,
            env=env if env is not None else self.subenv,
            cwd=cwd,
        )

    def check_output(  # noqa: PLR0913  # subprocess.check_output() takes many parameters
        self,
        cmd: PathSeq,
        *,
        shell: bool = False,
        env: dict[str, str] | None = None,
        cwd: pathlib.Path | None = None,
        encoding: str = "UTF-8",
        bufsize: int = 0,
    ) -> str:
        """Run a child process in a UTF-8 locale."""
        return subprocess.check_output(
            cmd,
            shell=shell,
            env=env if env is not None else self.subenv,
            cwd=cwd,
            encoding=encoding,
            bufsize=bufsize,
        )

    def Popen(  # noqa: N802,PLR0913  # subprocess.check_output() takes many parameters
        self,
        cmd: PathSeq,
        *,
        shell: bool = False,
        env: dict[str, str] | None = None,
        cwd: pathlib.Path | None = None,
        encoding: str = "UTF-8",
        bufsize: int = 0,
        stdin: int | IO[str] | None = None,
        stdout: int | IO[str] | None = None,
        stderr: int | IO[str] | None = None,
    ) -> Popen:
        """Run a child process in a UTF-8 locale."""
        return subprocess.Popen(
            cmd,
            shell=shell,
            env=env if env is not None else self.subenv,
            cwd=cwd,
            encoding=encoding,
            bufsize=bufsize,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
        )


@dataclasses.dataclass(frozen=True)
class Package:
    """A single package found in a branch."""

    name: str
    branch: str
    checksums: dict[str, pathlib.Path]
    overlays: list[pathlib.Path]

    def __str__(self) -> str:
        """Provide a human-readable package description."""
        return (
            f"package {self.name} branch {self.branch} distros ("
            + ", ".join(sorted(self.checksums.keys()))
            + ")"
        )


REMOTE_HOSTS: Final = {
    "centos/7": "spbuild@pkg-builder-rpm.sof-vm.storpool.local",
    "centos/8": "spbuild@pkg-builder-rpm.sof-vm.storpool.local",
    "centos/9": "spbuild@pkg-builder-rpm.sof-vm.storpool.local",
    "debian/unstable": "spbuild@pkg-builder-deb.sof-vm.storpool.local",
    "debian/bookworm": "spbuild@pkg-builder-deb.sof-vm.storpool.local",
    "debian/buster": "spbuild@pkg-builder-deb.sof-vm.storpool.local",
    "debian/bullseye": "spbuild@pkg-builder-deb.sof-vm.storpool.local",
    "debian/stretch": "debian9@d9build",
    "ubuntu/bionic": "spbuild@pkg-builder-deb.sof-vm.storpool.local",
    "ubuntu/focal": "spbuild@pkg-builder-deb.sof-vm.storpool.local",
    "ubuntu/impish": "spbuild@pkg-builder-deb.sof-vm.storpool.local",
    "ubuntu/jammy": "spbuild@pkg-builder-deb.sof-vm.storpool.local",
    "ubuntu/lunar": "spbuild@pkg-builder-deb.sof-vm.storpool.local",
}

_VARIANTS_IN_ORDER: Final = [
    var for var in variant.get_all_variants_in_order() if var.builder.branch in REMOTE_HOSTS
]

BRANCHES: Final = {
    var.builder.branch: Branch(
        name=var.builder.branch,
        flavor="centos" if var.family == "redhat" else "debian",
        repo_id=var.name,
    )
    for var in _VARIANTS_IN_ORDER
}

ORDERED_BRANCHES: Final = [BRANCHES[var.builder.branch] for var in _VARIANTS_IN_ORDER]

assert len(ORDERED_BRANCHES) == len(BRANCHES), repr(  # noqa: S101  # We really want this
    (
        ORDERED_BRANCHES,
        BRANCHES,
        len(ORDERED_BRANCHES),
        len(BRANCHES),
        [branch for branch in ORDERED_BRANCHES if branch.name not in BRANCHES],
    )
)
