# SPDX-FileCopyrightText: StorPool <support@storpool.com>
# SPDX-License-Identifier: BSD-2-Clause
"""Repository-related actions."""

from __future__ import annotations

import subprocess
import sys
import typing

from .scan import deb as s_deb
from .scan import rpm as s_rpm
from .scan import yum as s_yum


if typing.TYPE_CHECKING:
    import pathlib
    from typing import Final

    from . import defs
    from . import scan


MARKER_FILE: Final = "sp-pkg-stash.txt"


def stash_package(cfg: defs.Config, built: scan.Built, repodir: pathlib.Path) -> None:
    """Stash a single package into a quasi-repository."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    marker_file: Final = repodir / MARKER_FILE
    if not marker_file.is_file():
        sys.exit(f"No {marker_file}; is {repodir} a quasi-repo?")

    dstdir: Final = repodir / cfg.branch.name / built.name / built.version
    dstdir.mkdir(mode=0o755, parents=True, exist_ok=True)
    subprocess.check_call(
        [
            "rsync",
            "-av",
            "--delete",
            "--",
            str(built.path) + "/",
            str(dstdir) + "/",
        ],
        shell=False,
    )


def reprepro_include(cfg: defs.Config, *, stash: bool = False) -> None:
    """Add a package to a reprepro repository."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    assert cfg.path.repo is not None  # noqa: S101  # mypy needs this
    assert cfg.pkgname is not None  # noqa: S101  # mypy needs this

    repodir: Final = cfg.path.repo / cfg.branch.flavor
    cfg.diag(f"Adding {cfg.pkgname} to the {repodir} reprepro repository")

    srcdir: Final = cfg.path.project / "build" / cfg.branch.name / cfg.pkgname
    if not srcdir.is_dir():
        sys.exit(f"No {srcdir} directory")

    built: Final = s_deb.scan_built(cfg, srcdir)

    if stash:
        stash_package(cfg, built, cfg.path.repo)
        return

    vers: Final = cfg.branch.name.split("/", maxsplit=1)[-1]
    cfg.diag(f"- making sure {repodir} has a {vers} branch")
    if not (repodir / "dists" / vers).is_dir():
        sys.exit(f"No {vers} branch in the {repodir} repository")

    cfg.diag(f"- removing a {cfg.pkgname} source package if it exists")
    cfg.check_call(
        ["reprepro", "--", "removesrc", vers, cfg.pkgname],
        cwd=repodir,
    )

    cfg.diag(f"- adding {built.changes}")
    cfg.check_call(
        ["reprepro", "--", "include", vers, built.changes],
        cwd=repodir,
    )


def remove_rpm_source(repo: s_yum.Repo, srcfile: s_rpm.RPMSource) -> list[list[str]]:
    """Output the commands needed to remove a file and stuff from stuff."""
    pkg: Final = repo.sources.repomd.primary.packages.get(srcfile.name)
    if pkg is None:
        return []

    res: Final = [["rm", "-fv", "--", str(pkg.path)]]
    srcname: Final = pkg.path.name

    for arch in repo.arches.values():
        for pkg in arch.repomd.primary.packages.values():
            if pkg.source != srcname:
                continue
            res.append(["rm", "-fv", "--", str(pkg.path)])

    return res


def examine_rpm(repo: s_yum.Repo, built: s_rpm.RPMBuilt) -> list[list[str]]:
    """Scan packages, build the list of commands."""
    srcfile, binaries = built.source, built.binaries

    cmds: Final = remove_rpm_source(repo, srcfile)

    rpath: Final = repo.sources.path / "Sources"
    cmds.append(["cp", "-vi", "--", str(srcfile.path), str(rpath) + "/"])
    for pkg in binaries:
        if pkg.arch != "noarch":
            rpath_pkgs = repo.arches[pkg.arch].path / "Packages"
            cmds.append(["cp", "-vi", "--", str(pkg.path), str(rpath_pkgs) + "/"])
            continue

        for arch in s_yum.ARCH_VALID:
            if arch not in repo.arches:
                continue
            rpath_pkgs = repo.arches[arch].path / "Packages"
            cmds.append(["cp", "-vi", "--", str(pkg.path), str(rpath_pkgs) + "/"])

    cmds.append(["createrepo_c", "--", str(repo.sources.path)])
    cmds.extend(["createrepo_c", "--", str(archrepo.path)] for archrepo in repo.arches.values())

    return cmds


def crepo_include(cfg: defs.Config, *, stash: bool = False) -> None:
    """Add a package to a yum repository."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    assert cfg.path.repo is not None  # noqa: S101  # mypy needs this
    assert cfg.pkgname is not None  # noqa: S101  # mypy needs this
    cfg.diag(f"Adding {cfg.pkgname} to the {cfg.path.repo} Yum repository")

    srcdir: Final = cfg.path.project / "build" / cfg.branch.name / cfg.pkgname
    if not srcdir.is_dir():
        sys.exit(f"No {srcdir} directory")

    built: Final = s_rpm.scan_built(cfg, srcdir)
    cfg.diag(
        f"- found a {built.source.name} source package and "
        f"{len(built.binaries)} binary packages"
    )

    sp_key_id: Final = s_rpm.find_sp_key(cfg)
    unsigned: Final = s_rpm.verify_sigs(cfg, built, key_id=sp_key_id)
    if unsigned:
        if any(path.parent != srcdir for path in unsigned):
            sys.exit(f"All RPM files should live in {srcdir}: {unsigned!r}")
        names: Final = [path.name for path in unsigned]
        cfg.diag(f"- invoking rpmsign on {names}")
        cfg.check_call(
            ["rpmsign", "--addsign", "--", *names],
            cwd=srcdir,
        )
        remaining: Final = s_rpm.verify_sigs(cfg, built, key_id=sp_key_id)
        if remaining:
            sys.exit(f"Unsigned files even after rpmsign: {remaining}")

    if stash:
        stash_package(cfg, built, cfg.path.repo)
        return

    repo: Final = s_yum.Repo(cfg.path.repo / cfg.branch.name)
    cmds: Final = examine_rpm(repo, built)
    cfg.diag(f"- running {len(cmds)} commands")
    for cmd in cmds:
        cfg.diag("  - " + " ".join(cmd))
        cfg.check_call(cmd)

    cfg.diag("- all done!")


def cmd_include(cfg: defs.Config) -> None:
    """Include a package into a repository."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    if cfg.branch.flavor == "debian":
        reprepro_include(cfg)
    else:
        crepo_include(cfg)


def cmd_stash(cfg: defs.Config) -> None:
    """Stash a package into a quasi-repository."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    if cfg.branch.flavor == "debian":
        reprepro_include(cfg, stash=True)
    else:
        crepo_include(cfg, stash=True)
